#
# Cylindrical Algebraic Decomposition
# Cell object
# License: GNU General Public License 3
# Uray M. János
# 2019-2020
#

import itertools;

load("poly.sage");
load("rootof.sage");
load("formula.sage");

class CADCell:
	"""
	Cell of the cylindrical algebraic decomposition, which may itself contain
	recursively more cells. This is the return type of `cylindrical_algebraic_-
	decomposition()` when a list of polynomials is given, representing the whole
	decomposition.
	"""
	
	#
	#  Basic
	# -------
	#
	
	def cells (self):
		"""Enumerate all (sub)cells of the decomposition."""
		if self._cells:
			for subcell in self._cells:
				for cell in subcell.cells():
					yield cell;
		else:
			yield self;
	
	def sample (self):
		"""Return one point of the cell."""
		return tuple(AA(value) for (value, var, exact) in self._sample);
	
	def dimension (self):
		"""Return the dimension of the cell."""
		return sum(1 for (value, var, exact) in self._sample if not exact);
	
	def formula (self, type=None):
		"""Return a CAD formula describing this cell."""
		parents = [];
		parent = self;
		while parent:
			parents.append(parent);
			parent = parent._parent;
		parents.reverse();
		def eval (cell):
			level = len(cell._sample);
			if level >= len(parents):
				return True;
			elif parents[level] == cell:
				return True if level == len(parents) - 1 else None;
			else:
				return False;
		return parents[0]._to_formula(type, eval);
	
	
	#
	# Private members
	# ---------------
	#
	# Members for any cell:
	# self._parent: the parent CADCell object (for root: None)
	# self._sample: a list containing a sample point in this cell,
	#    with further information. Each element corresponds to a coordinate,
	#    which is a tuple: (value, var, exact), where:
	#    value: the value of the sample coordinate
	#    var: the polynomial variable whose value is given
	#    exact: is the coordinate an exact root (not in an interval)? (True/False)
	# self._formula: the input formula substituted in self._sample,
	#    if the CAD is built for a formula.
	#    It can be either a Logic or an OptimizedFormula.
	# self._value: stores the truth value (True/False) of self._formula if
	#    the whole cell has a constant value (otherwise None).
	# self._cells: the list of subcells as CADCell objects, or None for leaf
	# self._watch: the Stopwatch object measuring the performance
	#
	# Members for non-leaf cells only:
	# self._var: next splitting variable (along the next axis)
	#    as the generator of the specific polynomial ring
	# self._quan: the quantifier of this variable, or None for free variable
	# self._field: common number field of the sample values in self._sample
	#    (created in self._common_field())
	# self._mpolys: list of multivariate polynomials whose last variable
	#    (which is self._var) splits the cell along the next axis.
	#    These polynomials come from the parameter of _split(), which comes
	#    from the projection or the input.
	# self._polys: univariate polynomials sampled from self._mpolys
	#    by the values in self._sample.
	# self._roots: list of ordered unique roots of self._polys.
	#    It contains 3-tuples: (root value, index of polynomial, index of root)
	# self._points: list of sample points, containing alternately the roots and
	#    one rational (possibly integer) value in each interval
	#
	
	# Create the CADCell object as a leaf (but later can be split)
	def __init__ (self, parent, sample, formula=None, watch=None):
		self._parent = parent;
		self._sample = sample;
		self._formula = formula;
		self._value = bool(formula) if formula != None and formula.is_constant() else None;
		self._cells = None;
		self._watch = watch;
	
	
	#
	# Building the CAD
	# ----------------
	#
	
	# Split the cell along the next axis
	# var: the splitting new variable
	# mpolys: list of (multivariate) polynomials to split along
	# order_inv: is order-invariance needed? (for McCallum's projection operator)
	# quan: the quantifier for the splitting variable, or None
	def _split (self, var, mpolys, order_inv=False, quan=None):
		self._var = var;
		self._quan = quan;
		self._mpolys = mpolys;
		self._field = AA;
		#if self._watch: self._watch.start("number field");
		#self._common_field();
		if self._watch: self._watch.start("substitution");
		self._substitute_sample(order_inv);
		if self._watch: self._watch.start("root finding");
		self._collect_roots();
		if self._watch: self._watch.start("sampling");
		self._create_sample_points();
		if self._watch: self._watch.start("subcells");
		self._create_subcells();
		if self._watch: self._watch.stop();
	
	# Bring self._sample to a common number field instead of individual
	# algebraic numbers. The field is saved in self._field.
	# It is assumed that the parent did the same, i.e. all elements but the
	# last are already in a common number field, namely self._parent._field.
	def _common_field (self):
		if not self._sample:
			self._field = QQ;
			return;
		last = self._sample[-1][0];
		try:
			last = QQ(last);
			self._field = self._parent._field;
		except:
			pgen = self._parent._field.gen();
			self._field, [pgen, last], _ = number_field_elements_from_algebraics([pgen, last], minimal=True, embedded=True);
			for i in range(len(self._sample) - 1):
				p, v, ex = self._sample[i];
				if p.parent() != QQ:
					p = p.polynomial()(pgen);
					self._sample[i] = (p, v, ex);
		_, v, ex = self._sample[-1];
		self._sample[-1] = (last, v, ex);
	
	# Create self._polys from self._mpolys by substitution.
	# order_inv: if True, the result is modified according to McCallum's algorithm
	# Used by _split().
	def _substitute_sample (self, order_inv=False):
		sample_values = tuple(p for p, v, ex in reversed(self._sample));
		alg_ring = self._field[self._var];
		rat_ring = QQ[self._var];
		def subs (poly):
			if sample_values:
				coeffs = [cf(sample_values) for cf in poly];
				if all(cf in QQ for cf in coeffs):
					return rat_ring(coeffs);
				else:
					return alg_ring(coeffs);
			else:
				return poly;
		self._polys = [subs(poly) for poly in self._mpolys];
		if order_inv:
			first = True;
			for i in range(len(self._polys)):
				if self._polys[i].is_zero():
					if self.dimension() == 0:
						# find a delineating polynomial
						for der in self._gen_partial(self._mpolys[i]):
							poly = subs(der);
							if not poly.is_zero():
								if first:
									self._mpolys = self._mpolys[:];
									first = False;
								self._mpolys[i] = der;
								self._polys[i] = poly;
								break;
					else:
						raise ValueError("polynomial vanishes identically on a {}-dim cell".format(self.dimension()));
	
	# Generate partial derivatives of poly of increasing degree.
	# Used by _substitute_sample() for order-invariant output.
	@staticmethod
	def _gen_partial (poly):
		ring = poly.parent();
		vars = [];
		while isinstance(ring.gen(), Polynomial):
			vars.append(ring.gen());
			ring = ring.base();
		polys = [poly];
		idx = 0;
		while True:
			for v in vars:
				dpoly = polys[idx].diff(v);
				yield dpoly;
				polys.append(dpoly);
			idx += 1;
	
	# Create self._roots from self._polys by merging their roots.
	# Used by _split().
	def _collect_roots (self):
		self._roots = [];
		for i, p in enumerate(self._polys):
			#proots = find_algebraic_roots(p);
			proots = p.roots(ring=AA, multiplicities=False);
			for j, r in enumerate(proots):
				if r in QQ:
					r = QQ(r) if r not in ZZ else ZZ(r);
				self._roots.append((r, i, j));
		self._roots.sort();
		self._remove_duplicate_roots();
	def _remove_duplicate_roots (self):
		self._roots = [next(rg) for r, rg in itertools.groupby(self._roots, lambda t : t[0])];
	
	# Create self._points from self._roots by finding good sample points.
	# Used by _split().
	def _create_sample_points (self):
		self._points = [];
		for a, b in self._iterate_root_intervals():
			if a != -Infinity:
				self._points.append(a);
			if a < 0 and b > 0:
				c = 0;
			elif a == -Infinity:
				c = (b.floor() if b not in ZZ else ZZ(b) - 1);
			else:
				c = (a.ceil() if a not in ZZ else ZZ(a) + 1);
				u = 1;
				while c >= b:
					u /= 2;
					if c - u > a:
						c -= u;
			self._points.append(c);
	
	# Iterate over intervals between self._roots, including infinite ends.
	# Used by _create_sample_points().
	def _iterate_root_intervals (self):
		if len(self._roots) == 0:
			yield (-Infinity, +Infinity);
			return;
		yield (-Infinity, self._roots[0][0]);
		for i in range(len(self._roots) - 1):
			yield (self._roots[i][0], self._roots[i+1][0]);
		yield (self._roots[-1][0], +Infinity);
	
	# Create self._cells from self._points, using self as base.
	# Used by _split().
	def _create_subcells (self):
		self._cells = [];
		exact = False;
		v = self._var;
		for p in self._points:
			sample = self._sample + [(p, v, exact)];
			formula = self._formula.subs({v : p}) if self._formula != None else None;
			self._cells.append(CADCell(self, sample, formula, watch=self._watch));
			exact = not exact;
	
	
	#
	# Formula construction
	# --------------------
	#
	
	# Search for constant truth value, and set self._value.
	def _combine_values (self):
		if self._cells == None: return;
		if self._quan:
			qe = self._quan.evaluator();
			for cell in self._cells:
				cell._combine_values();
				assert cell._value is not None;
				qe.add(cell._value);
				self._value = qe.value();
				if self._value is not None:
					return;
			qe.finish();
			self._value = qe.value();
		else:
			for cell in self._cells:
				cell._combine_values();
			self._value = self._cells[0]._value;
			for cell in self._cells:
				if cell._value != self._value:
					self._value = None;
					break;
	
	# Return a formula for the True-valued subcells.
	# type: "rooty" or "poly", see the functions below
	# eval: if given, determines the value of the cells (True/False/None).
	#    Otherwise it is determined by the input formula.
	def _to_formula (self, type=None, eval=None):
		if type is None: type = "rooty";
		if eval is None: eval = lambda cell : cell._value;
		if type == "rooty":
			return self._to_rooty_formula(eval);
		elif type == "poly":
			return self._to_poly_formula(eval);
		else:
			raise ValueError("invalid formula type '" + type + "'");
	
	# Return a formula for the True-valued subcells.
	# The formula may contain rootof()-expressions.
	def _to_rooty_formula (self, eval=None):
		value = eval(self);
		if value != None:
			return logic(value);
		formulas = self._gen_subformulas(CADCell._to_rooty_formula, eval);
		same_formula_ranges = [];
		last = 0;
		for i in range(len(self._cells)):
			if i == len(formulas)-1 or formulas[i+1] != formulas[last]:
				same_formula_ranges.append((last, i));
				last = i+1;
		v = SR(self._var);
		def get_root (i):
			root = self._roots[i];
			return rootof(self._mpolys[root[1]], v, root[2]);
		subforms = [];
		for [low_cell, upp_cell] in same_formula_ranges:
			subform = logic_true;
			low_root = (low_cell-1) // 2;
			upp_root = upp_cell // 2;
			if low_root == upp_root:
				subform &= (v == get_root(low_root));
			else:
				if low_root >= 0:
					if low_cell % 2 == 0:
						subform &= (v > get_root(low_root));
					else:
						subform &= (v >= get_root(low_root));
				if upp_root < len(self._roots):
					if upp_cell % 2 == 0:
						subform &= (v < get_root(upp_root));
					else:
						subform &= (v <= get_root(upp_root));
			subform &= formulas[low_cell];
			subforms.append(subform);
		return logic_or(subforms);
	
	# Return a formula for the True-valued subcells.
	# The formula contains only polynomials.
	def _to_poly_formula (self, eval=None):
		value = eval(self);
		if value != None:
			return logic(value);
		def subset (mask0, mask1):
			return (mask0 & (mask1 ^^ -1)) == 0;
		formulas = self._gen_subformulas(CADCell._to_poly_formula, eval);
		sign_masks = [];
		for p in self._points:
			sign_bits = 0;
			for i, f in enumerate(self._polys):
				sign_bits |= sign_to_mask(f(p)) << 3*i;
			sign_masks.append(sign_bits);
		polys = [SR(f) for f in self._mpolys];
		v = SR(self._var);
		def mask_to_formula (mask):
			mask ^^= -1;
			formula = logic_true;
			while mask:
				lowest_bit_idx = (mask & -mask).nbits() - 1;
				index = lowest_bit_idx // 3;
				if index >= len(polys): break;
				sign_mask = (mask >> index*3) & 7;
				form = mask_to_relation(7 ^^ sign_mask, polys[index]);
				form -= form.lhs()({v: 0});
				formula &= form;
				mask &= (7 << index*3) ^^ -1;
			return formula;
		def find_formula_mismatch (mask):
			last = None;
			for i in range(len(self._cells)):
				if not subset(sign_masks[i], mask):
					continue;
				if last is not None and formulas[i] != formulas[last]:
					return last, i;
				last = i;
			return last, None;
		def find_splitting_cond (signs0, signs1, mask):
			for i, f in enumerate(polys):
				bits = 7 << 3*i;
				s0, s1 = signs0 & bits, signs1 & bits;
				if s0 != s1:
					if s0 >> 3*i == 2:
						ms = (s1 ^^ bits, s1);
					else:
						ms = (s0, s0 ^^ bits);
					possible_signs = union_in(mask);
					return i, tuple((m & possible_signs) | (bits ^^ -1) for m in ms);
			else:
				return None, None;
		def union_in (mask):
			res = 0;
			for signs in sign_masks:
				if subset(signs, mask):
					res |= signs;
			return res;
		def stronger_cond (cond0, cond1):
			if subset(cond0, cond1):
				return True;
			for signs in sign_masks:
				if subset(signs, cond0) and not subset(signs, cond1):
					return False;
			else:
				return True;
		def partial_formula (mask, base_cond, base_form):
			i, j = find_formula_mismatch(mask);
			if i is None:
				return logic_false;
			elif j is None:
				return logic_and(base_form, formulas[i]);
			else:
				pos, conds = find_splitting_cond(sign_masks[i], sign_masks[j], mask);
				if pos is None:
					raise NotImplementedError("CAD cannot be expressed using only the input polynomials, and adding new polynomials is not yet implemented");
				else:
					formula = logic_false;
					for cond in conds:
						form = mask_to_formula(cond);
						subform = partial_formula(mask & cond, cond, form);
						if not stronger_cond(cond, base_cond):
							subform = logic_and(base_form, subform);
						formula |= subform;
					return simplify_logic(formula);
		return partial_formula(-1, -1, logic_true);
	
	# Return a formula for each subcells in a list, using 'method',
	# which is either of the functions above.
	def _gen_subformulas (self, method, eval):
		formulas = [method(cell, eval) for cell in self._cells];
		for i in range(1, len(self._cells), 2):
			cell = self._cells[i];
			if cell._is_formula_ok(formulas[i-1], eval) and self._formula_complexity(formulas[i-1]) <= self._formula_complexity(formulas[i]):
				formulas[i] = formulas[i-1];
			elif cell._is_formula_ok(formulas[i+1], eval) and self._formula_complexity(formulas[i+1]) <= self._formula_complexity(formulas[i]):
				formulas[i] = formulas[i+1];
		return formulas;
	
	# Return whether the formula describes this cell exactly.
	def _is_formula_ok (self, formula, eval):
		try:
			formula = formula({v : p for p, v, _ in self._sample});
			value = eval(self);
			if value != None:
				return formula == value;
			for cell in self._cells:
				if not cell._is_formula_ok(formula, eval):
					return False;
			return True;
		except ValueError: # formula is invalid here, e.g. div by zero or rootof() index
			return False;
	
	# Return the complexity of a formula.
	# Each relational atom costs 1 (=,<,>) or 2 (<=,>=,!=).
	@staticmethod
	def _formula_complexity (formula):
		c = 0;
		for atom in formula.atoms():
			if atom.is_constant():
				pass;
			elif atom.operator() in (operator.eq, operator.lt, operator.gt):
				c += 1;
			elif atom.operator() in (operator.le, operator.ge, operator.ne):
				c += 2;
		return c;

