#
# Cylindrical Algebraic Decomposition
# (basic + projection)
# License: GNU General Public License 3
# Uray M. János
# 2019-2020
#

import itertools;
from sage.rings.polynomial.multi_polynomial import MPolynomial;

load("poly.sage");
load("cad-cell.sage");
load("formula.sage");
load("qepcad.sage");


def cylindrical_algebraic_decomposition (input, **kwargs):
	"""
	Calculate the cylindrical algebraic decomposition.
	The input parameter can be any of the following:
	1. (List of) multivariate polynomials. In this case, the function returns
		a full CAD, which is an instance of the CADCell class.
	2. A symbolic system of inequalities as a logical formula (instance of
		Logic, or a relational expression), and an optional vars parameter,
		which is a tuple of the variables, describing the variable order for CAD.
		In this case, the function returns an equivalent quantifier-free CAD formula.
	Optional keyword arguments:
	`vars`: the order of the variables (tuple)
	`proj`: the projection operator (see `cad_projection_by_name()`)
	Optional keyword arguments for formula conversion:
	`algorithm`: "builtin" or "qepcad"
	`output`: the type of the output formula
		"rooty": may contain `rootof()` expressions
		"poly": contains only polynomials
	"""
	polys = formula = None;
	if isinstance(input, list):
		polys = input;
	elif isinstance(input, Polynomial):
		polys = [input];
	elif isinstance(input, MPolynomial):
		polys = [input];
	elif isinstance(input, Expression):
		if input.is_relational():
			formula = logic(input);
		else:
			polys = [input];
	elif isinstance(input, Logic):
		formula = input;
	if polys is not None:
		return cad_polys_to_cells(polys, **kwargs);
	elif formula is not None:
		return cad_formula_reduce(formula, **kwargs);
	else:
		raise TypeError("cylindrical_algebraic_decomposition() expects either a list of polynomials or a logical formula");


def cad_formula_reduce (formula, vars=None, algorithm="builtin", output=None, proj=None, watch=None):
	"""
	Convert a logical formula to a simplier and quantifier-free one by CAD.
	
	`formula`: the input formula
	`vars`: the variable ordering for CAD (tuple)
	`algorithm`: "builtin" or "qepcad"
	`output`: the type of the output formula
		"rooty": may contain `rootof()` expressions
		"poly": contains only polynomials
	`proj`: the projection operator (see `cad_projection_by_name()` below)
	`watch`: the `Stopwatch` object to measure performance
	"""
	if watch: watch.start("preprocess");
	formula = convert_formula_to_poly(formula);
	formula = formula.to_prenex();
	free_vars = formula.free_variables();
	base_formula = formula.quantifier_base();
	quantifiers = formula.top_quantifiers();
	quan_vars = tuple(quan.variable() for quan in quantifiers);
	if vars is not None and set(vars) & set(quan_vars):
		raise ValueError("variable list must not contain quantified variable");
	if vars is None:
		vars = free_vars;
	vars += quan_vars;
	ring = cad_input_to_poly_ring(base_formula, vars);
	if watch: watch.stop();
	if algorithm == "builtin":
		if watch: watch.start("preprocess");
		polys = PolyFactorList(ring);
		opt_formula = OptimizedFormula.create(base_formula, polys);
		if watch: watch.stop();
		cad = cad_polys_to_cells(polys, proj=proj, formula=opt_formula, quantifiers=quantifiers, watch=watch);
		print "cells:", sum(1 for cell in cad.cells()); # TEST
		if watch: watch.start("formula");
		cad._combine_values();
		output = cad._to_formula(type=output);
		if watch: watch.stop();
		return output;
	elif algorithm == "qepcad":
		if formula.is_constant():
			return formula;
		if watch: watch.start("conversion");
		input = formula_to_qepcad(formula);
		polys = PolyFactorList(ring);
		vars = polys.variables();
		if watch: watch.start("QEPCAD");
		output = use_qepcad(input, vars, len(free_vars));
		if watch: watch.start("conversion");
		output = qepcad_to_formula(output);
		if watch: watch.stop();
		return output;
	else:
		raise ValueError("invalid algorithm \"" + algorithm + "\"");

def cad_polys_to_cells (polys, vars=None, proj=None, formula=None, quantifiers=None, watch=None):
	"""
	Build CAD for list of polynomials.
	
	`polys`: list of polynomials
	`vars`: variable order (tuple)
	`proj`: the projection operator (see `cad_projection_by_name()`)
	`formula`, `quantifiers`: used internally by `cad_formula_reduce()`
	`watch`: the `Stopwatch` object to measure performance
	"""
	project = cad_projection_by_name(proj);
	order_inv = cad_project_needs_order_inv(project);
	if watch: watch.start("projection");
	if not isinstance(polys, PolyFactorList):
		list = PolyFactorList(cad_input_to_poly_ring(polys, vars));
		for f in polys:
			list.add(f);
		polys = list;
	vars = polys.variables();
	quans = quantifiers;
	polys_levels = [polys];
	for i in range(1, len(vars)):
		polys = project(polys);
		polys_levels.append(polys);
	for polys in polys_levels:
		polys.filter(lambda poly : poly.degree() >= 1);
		#print "proj:", polys._list; # TEST
	print "proj:", [len(polys) for polys in polys_levels]; # TEST
	if watch: watch.stop();
	root = CADCell(None, [], formula, watch=watch);
	for i, v in enumerate(vars):
		polys = polys_levels.pop();
		quan = quans[i - len(vars)] if quans and i >= len(vars) - len(quans) else None;
		if order_inv and i == len(vars) - 1:
			order_inv = False;
		for cell in root.cells():
			if cell._value == None:
				cell._split(v, polys, order_inv=order_inv, quan=quan);
	return root;


# Calculate the recursive polynomial ring of the CAD input.
# input: either a logical formula containing only polynomials
#    (after convert_formula_to_poly()), or a list of polynomials.
# vars: the input variable order for CAD, if given.
def cad_input_to_poly_ring (input, vars=None):
	def get_input_vars_set (input):
		if isinstance(input, Logic):
			return set(input.variables());
		else:
			vars = set();
			for f in input:
				vars.update(SR(f).variables());
			return vars;
	
	def sort_vars (input_vars, vars):
		if vars is None:
			from sage.symbolic.comparison import print_sorted;
			return tuple(reversed(print_sorted(input_vars)));
		else:
			vars_set = frozenset(vars);
			if len(vars_set) < len(vars):
				raise ValueError("duplicate variable in vars parameter");
			if not input_vars <= vars_set:
				raise ValueError("vars parameter is incomplete");
			return tuple(vars);
	
	def only_rational (input):
		if isinstance(input, Logic):
			args = input.operands();
		elif isinstance(input, Expression):
			if input.is_constant():
				return input in QQ;
			elif input.is_symbol():
				return True;
			else:
				args = input.iterator();
		else:
			args = input;
		return all(only_rational(arg) for arg in args);
	
	input_vars = get_input_vars_set(input);
	vars = sort_vars(input_vars, vars);
	ring = QQ if only_rational(input) else AA;
	for v in vars:
		ring = ring[v];
	return ring;


#
# Projection
# ----------
#

def cad_projection_by_name (name):
	if name == None: return cad_collins_project_2;
	elif name == "collins":  return cad_collins_project;
	elif name == "collins2": return cad_collins_project_2;
	elif name == "mccallum": return cad_mccallum_project;
	else:
		raise ValueError("invalid projection operator \"" + name + "\"");

def cad_project_needs_order_inv (func):
	return func == cad_mccallum_project;

# Collins' projection operator
def cad_collins_project (polys):
	proj_polys = PolyFactorList(polys.ring().base());
	red_polys = [];
	for f in polys:
		g = f;
		while g.degree() >= 0:
			proj_polys.add(g.lc());
			red_polys.append(g);
			g -= g.lt();
	for f in red_polys:
		for psc in subresultant(f, f.diff(), ret="psc"):
			proj_polys.add(psc);
	for f, g in itertools.combinations(red_polys, 2):
		for psc in subresultant(f, g, ret="psc"):
			proj_polys.add(psc);
	return proj_polys;

# Collins' reduced projection operator
def cad_collins_project_2 (polys):
	d = polys.dimension();
	proj_polys = PolyFactorList(polys.ring().base());
	red_polys = [];
	for f in polys:
		k = finite_common_zeros(list(reversed(f.coefficients())), d-1);
		g = f;
		while k >= 0:
			proj_polys.add(g.lc());
			red_polys.append(g);
			g -= g.lt();
			k -= 1;
	for f in red_polys:
		pscs = subresultant(f, f.diff(), ret="psc");
		k = finite_common_zeros(pscs, d-1);
		for i, psc in enumerate(pscs):
			proj_polys.add(psc);
			if i == k:
				break;
	for f, g in itertools.combinations(red_polys, 2):
		pscs = subresultant(f, g, ret="psc");
		k = finite_common_zeros(pscs, d-1);
		for i, psc in enumerate(pscs):
			proj_polys.add(psc);
			if i == k:
				break;
	return proj_polys;

# McCallum's projection operator
def cad_mccallum_project (polys):
	proj_polys = PolyFactorList(polys.ring().base());
	for f in polys:
		cfs = list(reversed(f.coefficients()));
		k = finite_common_zeros(cfs, polys.dimension() - 1);
		for i in range(min(k+1, len(cfs))):
			proj_polys.add(cfs[i]);
	for f in polys:
		if f.degree() > 0:
			proj_polys.add(f.resultant(f.diff()));
	for f, g in itertools.combinations(polys, 2):
		proj_polys.add(f.resultant(g));
	return proj_polys;


# Determine how many of the given polynomials are sufficient to ensure that
# they have only a finite number of common roots.
# (Auxiliary function for projection operators.)
# polys: a list of polynomials (of type QQ[x][y]...)
# dim: the number of variables
# The return value is a number n which means that the first n+1 polynomials
# are proven to have only finite number of common roots (possibly none).
# If n == len(polys), it means that no such number was found.
def finite_common_zeros (polys, dim):
	def swap_last_two_vars (f):
		ring = f.parent();
		v1 = ring.gen();
		v2 = ring.base().gen();
		cf = f.coefficients(sparse=False);
		return ring(sum(g(v1)*v2^i for i, g in enumerate(cf)));
	if not polys or dim == 0: return 0;
	proj_f, proj_g = [], [];
	last_f, last_g = None, None;
	for i, f in enumerate(polys):
		if f.is_zero(): continue;
		if dim == 1: return i;
		if f.degree() == 0:
			proj_f.append(f.constant_coefficient());
		elif last_f is not None:
			proj_f.append(f.resultant(last_f));
		g = swap_last_two_vars(f);
		if g.degree() == 0:
			proj_g.append(g.constant_coefficient());
		elif last_g is not None:
			proj_g.append(g.resultant(last_g));
		if finite_common_zeros(proj_f, dim-1) < len(proj_f):
			if finite_common_zeros(proj_g, dim-1) < len(proj_g):
				return i;
		last_f, last_g = f, g;
	return len(polys);

