#
# Miscellanea for formulas
# License: GNU General Public License 3
# Uray M. János
# 2019-2020
#

from sage.symbolic.operators import add_vararg, mul_vararg;

load("logic.sage");

# 1:(-), 2:(0), 4:(+)
def sign_to_mask (value):
	return 1 << (sign(value) + 1);
def relation_to_mask (rel):
	if rel == operator.lt: return 1;
	if rel == operator.eq: return 2;
	if rel == operator.le: return 3;
	if rel == operator.gt: return 4;
	if rel == operator.ne: return 5;
	if rel == operator.ge: return 6;
def mask_to_relation (mask, expr):
	if mask == 0: return False;
	if mask == 1: return expr < 0;
	if mask == 2: return expr == 0;
	if mask == 3: return expr <= 0;
	if mask == 4: return expr > 0;
	if mask == 5: return expr != 0;
	if mask == 6: return expr >= 0;
	if mask == 7: return True;
def flip_sign_mask (mask):
	return ((mask & 4) >> 2) | (mask & 2) | ((mask & 1) << 2);


def simplify_logic (formula):
	"""
	Apply distributivity rule to simplify logical formula,
	e.g. convert (A /\\ B) \\/ (A /\\ C) to A /\\ (B \\/ C).
	"""
	out_op = formula.operator();
	if not (out_op == logic_and or out_op == logic_or):
		return formula;
	ops = formula.operands();
	def all_equal (iterator):
		first = None;
		for x in iterator:
			if first is None:
				first = x;
			else:
				if x != first:
					return None;
		else:
			return first;
	in_op = all_equal(op.operator() for op in ops);
	if not ((in_op == logic_and or in_op == logic_or) and in_op != out_op):
		return formula;
	pre = 0;
	min_len = min(len(op.operands()) for op in ops);
	while pre < min_len and all_equal(op.operands()[pre] for op in ops) is not None:
		pre += 1;
	suf = 0;
	while suf < min_len and all_equal(op.operands()[-suf-1] for op in ops) is not None:
		suf += 1;
	if pre == 0 and suf == 0:
		return formula;
	beg = pre;
	end = (-suf if suf != 0 else max(len(op.operands()) for op in ops));
	return in_op(
		ops[0].operands()[:beg] +
		[out_op(in_op(op.operands()[beg:end]) for op in ops)] +
		ops[0].operands()[end:]
	);


def convert_formula_to_poly (formula):
	"""
	Convert a logical formula to one that contains only polynomials and
	all right-hand-sides are zero.
	"""
	
	def convert_formula (form):
		"""
		Perform the actual conversion on nodes, as a parameter of
		Logic.recursive_replace().
		"""
		op = form.operator();
		if op in (logic_and, logic_or):
			return form;
		elif isinstance(op, Quantifier):
			return form;
		elif isinstance(form, LogicConstant):
			return form;
		elif isinstance(form, LogicRelation):
			form = SR(form);
			if form.rhs() != 0:
				form -= form.rhs();
			form = form.expand();
			res = convert_relation_step(form);
			if res is None:
				return logic(form);
			else:
				return convert_formula(logic(res));
		else:
			raise ValueError("formula {} cannot be polynomialized".format(form));
	
	def convert_relation_step (form):
		"""
		Perform one step of the conversion on relations.
		Return None if the relation is ok, otherwise return a modified version.
		It is to be called repeatedly until None is returned.
		"""
		subform = find_problem(form.lhs());
		if subform is None:
			return None;
		op = subform.operator();
		if op == operator.pow:
			exp = subform.op[1];
			if exp.is_integer() and exp < 0:
				return convert_fraction(form);
			if exp in QQ:
				return convert_sqrt(form, subform);
		elif op in (min_symbolic, max_symbolic):
			return convert_min_max(form, subform);
		elif op == sage.functions.other.abs:
			return convert_abs(form, subform);
		raise ValueError("expressions like {} are not supported by CAD".format(subform));
	
	def find_problem (expr):
		"""
		Find a non-polynomial part of the expression which must be dealt with,
		or return None if expression is a polynomial.
		"""
		op = expr.operator();
		if op is None:
			if expr.is_numeric() and expr.is_exact():
				return None;
			elif expr.is_symbol():
				return None;
		elif op in (add_vararg, mul_vararg):
			for arg in expr.iterator():
				f = find_problem(arg);
				if f is not None:
					return f;
			return None;
		elif op is operator.pow:
			exp = expr.op[1];
			if exp.is_integer() and exp >= 0:
				return find_problem(expr.op[0]);
		return expr;
	
	def convert_fraction (form):
		"""
		Convert a fractional relation to a polynomial formula.
		"""
		expr, rel = form.lhs(), form.operator();
		num, den = expr.numerator_denominator();
		if not den.is_numeric():
			if rel == operator.eq or rel == operator.ne:
				return logic_and(rel(num, 0), den != 0);
			else:
				return logic_or(
					logic_and(rel(num, 0), den > 0),
					logic_and(rel(-num, 0), den < 0)
				);
		elif not expr.is_trivially_equal(num):
			return rel(num, 0);
		return None;
	
	def convert_min_max (form, mform):
		"""
		Convert a relation containing a symbolic min/max to a polynomial formula.
		"""
		expr, rel = form.lhs(), form.operator();
		# Formulas like 2*max(x, y) + z > w (i.e. monotonic formulas)
		# can just be expanded to 2*x + z > w \/ 2*y + z > w.
		# Other formulas F(max(x, y)) are converted to
		# (x >= y /\ F(x)) \/ (y >= x /\ F(y)).
		monotonic = rel not in (operator.eq, operator.ne) and is_monotonic(mform, expr);
		max = (mform.operator() == max_symbolic);
		if monotonic:
			cases = [];
			for best in mform.iterator():
				cases.append(form.subs({mform : best}));
			if max ^^ bool(monotonic > 0) ^^ (rel in (operator.lt, operator.le)):
				return logic_and(cases);
			else:
				return logic_or(cases);
		else:
			cases = [];
			for best in mform.iterator():
				subforms = [];
				for other in mform.iterator():
					if other != best:
						subforms.append(best >= other if max else best <= other);
				subforms.append(form.subs({mform : best}));
				cases.append(logic_and(subforms));
			return logic_or(cases);
	
	def convert_abs (form, abs):
		"""
		Convert a relation containing a symbolic abs() to a polynomial formula.
		"""
		arg = abs.op[0];
		return form.subs({abs : max_symbolic(arg, -arg)});
	
	def is_monotonic (sub, expr):
		"""
		Try to decide whether an expression is monotonic in sub.
		+1: provably monotonically increasing
		-1: provably monotonically decreasing
		None: not sure, maybe not monotonic
		"""
		monot = None;
		for cf, n in expr.coefficients(sub):
			if n == 0: continue;
			if not cf.is_numeric() or not n.is_integer():
				return None;
			if ZZ(n) % 2 == 0:
				return None;
			sign = sgn(cf);
			assert sign == -1 or sign == +1;
			if monot is not None and sign != monot:
				return None;
			monot = sign;
		return monot;
	
	def convert_sqrt (form, root):
		"""
		Convert a relation containing an nth root to a polynomial formula.
		"""
		
		def is_root (expr, even=False):
			"""
			Is expr an nth root?
			even: check only even roots (otherwise all roots)
			"""
			if isinstance(expr, Expression) and expr.operator() == operator.pow:
				if not expr.op[1].is_integer() and expr.op[1] in QQ:
					if not even or root_exp(expr) % 2 == 0:
						return True;
			return False;
		
		def has_root (expr, even=False):
			"""
			Does expr contain an nth root?
			even: check only even roots (otherwise all roots)
			"""
			w0, w1 = SR.wild(0), SR.wild(1);
			if not isinstance(expr, Expression):
				return False;
			for sub in expr.find(w0^w1):
				if is_root(sub, even) or has_root(sub.op[0], even):
					return True;
			return False;
		
		def root_exp (root):
			"""Exponent of the root, e.g. 3 for x^(2/3)"""
			return QQ(root.op[1]).denom();
		
		def collect_root_factors (expr, even=False):
			"""Return a list of all factors of expr that are roots"""
			expr = expr.factor() if expr != 0 else expr;
			factors = expr.iterator() if isinstance(expr, Expression) and expr.operator() == mul_vararg else [expr];
			return [fac for fac in factors if is_root(fac, even)];
		
		expr, rel = form.lhs(), form.operator();
		# Try to write expr = a b^{1/n} + c d^{1/n} if possible,
		# where a and c is rootless, or if n is even, contains no even roots.
		even = has_root(expr, even=True);
		cfs = expr.coefficients(root, sparse=False);
		left, right = expr - cfs[0], cfs[0];
		l_roots = collect_root_factors(left, even=even);
		r_roots = collect_root_factors(right, even=even);
		all_roots = l_roots + r_roots;
		b, d = product(l_roots), product(r_roots);
		a, c = left / b, right / d;
		if has_root(a, even=even) or has_root(c, even=even):
			return convert_sqrt_general(form, root);
		n = gcd([QQ(root.op[1]) for root in all_roots]).denom();
		b, d = b ^ n, d ^ n;
		# Equivalent to a b^{1/n} + c d^{1/n}
		if n%2 == 1:
			return rel(a^n*b + c^n*d, 0);
		else:
			head = logic_and(root ^ root_exp(root) >= 0 for root in all_roots);
			if rel in (operator.eq, operator.ne):
				sign_ok = logic_or(logic_and(a >= 0, c <= 0), logic_and(a <= 0, c >= 0));
				body = logic_and(sign_ok, a^n*b == c^n*d);
				body |= logic_and(b == 0, d == 0);
			else:
				greater = operator.gt if rel in (operator.gt, operator.le) else operator.lt;
				agt0, cgt0 = greater(a, 0), greater(c, 0);
				body = logic_and(agt0, logic_or(logic_and(cgt0, logic_or(b > 0, d > 0)), a^n*b > c^n*d));
				body |= logic_and(logic_not(agt0), cgt0, a^n*b < c^n*d);
			if rel in (operator.ne, operator.le, operator.ge):
				body = logic_not(body);
			return head & body;
	
	def convert_sqrt_general (form, root):
		"""
		Fallback for convert_sqrt(). It uses a general method
		creating a new variable.
		"""
		v = new_variable();
		exp = QQ(root.op[1]).denom();
		vdef = logic(v^exp == root^exp);
		if exp % 2 == 0:
			vdef &= logic(v >= 0);
		return logic_exists(v, logic_and(vdef, form.subs({root : v})));
	
	new_vars = [];
	def new_variable ():
		"""Create a new variable"""
		def next_name (name):
			next = "";
			add = True;
			for ch in reversed(name):
				if add:
					if ch != "z":
						ch = chr(ord(ch) + 1);
						add = False;
					else:
						ch = "a";
				next = ch + next;
			if add:
				next = "a" + next;
			return next;
		vars = formula.variables();
		name = "a";
		var = SR.symbol(name);
		while var in vars or var in new_vars:
			name = next_name(name);
			var = SR.symbol(name);
		new_vars.append(var);
		return var;
	
	return formula.recursive_replace(convert_formula);


class OptimizedFormula:
	"""
	Optimized version of polynomial formula.
	It mimics an instance of `Logic` which contains only polynomials and no
	quantifiers. It optimizes its behaviour by using polynomial rings instead
	of general symbolic expressions, and also by factoring the polynomials
	and manipulating repeated factors only once.
	Its main function is to substitute a variable and simplify the formula,
	possibly to a constant true/false.
	It is used by `CADCell`, interchangably with the original `Logic` formula.
	"""
	
	@classmethod
	def create (cls, formula, polys):
		"""
		Convert a Logic formula to an OptimizedFormula.
		The formula must contain only polynomials, no quantifiers, and all
		relations must be reduced to zero on the right.
		`polys`: a new empty instance of `PolyFactorList` created for this
			formula. It determines the substitution order the variables in subs().
		   It is filled by all factors of polynomials occuring in the formula.
		"""
		def rec_process (form, unit=True):
			"""
			Implement this conversion recursively.
			`unit`: distinguish between `logic_and` and `logic_or`
			"""
			op = form.operator();
			if isinstance(op, Quantifier):
				raise TypeError("OptimizedFormula does not support quantified formula");
			elif op in (logic_and, logic_or):
				if unit == (op == logic_and):
					return [rec_process(sub, not unit) for sub in form.operands()];
				else:
					return [rec_process(form, not unit)]; # if outer op is wrong
			elif form.is_constant():
				return bool(form);
			elif isinstance(form, LogicRelation):
				if form.rhs():
					raise ValueError("OptimizedFormula requires zero right-hand-side");
				trace, const = polys.add(form.lhs());
				assert const != 0; # otherwise relation would be constant
				mask = relation_to_mask(form.operator());
				if const < 0:
					mask = flip_sign_mask(mask);
				return (trace, mask);
			else:
				raise TypeError("OptimizedFormula does not support formulas like {}".format(form));
		
		self = OptimizedFormula();
		self._tree = rec_process(formula);
		self._vars = polys.variables();
		self._polys = [f(*self._vars) for f in polys]; # reverse variables
		self._consts = 0;
		return self;
	
	def subs (self, value):
		"""
		Substitute the next variable to this formula, and return a new
		`OptimizedFormula`. The order of variables is determined by the polys
		argument of `create()`.
		"""
		subs_polys = self._polys[:];
		if isinstance(value, dict):
			for v in value:
				if v != self._vars[0]:
					raise ValueError("invalid variable for OptimizedFormula");
				value = value[v];
		done = [self._consts, self._consts]; # really two variables, [done, consts], but Python 2 lacks 'nonlocal'
		
		def rec_subs (form, unit=True):
			"""
			Implement this conversion recursively.
			`unit`: distinguish between `logic_and` and `logic_or`
			"""
			if isinstance(form, list):
				return rec_subs_chain(form, unit);
			elif isinstance(form, tuple):
				return subs_relation(form);
			elif isinstance(sub, bool):
				return sub;
		
		def rec_subs_chain (chain, unit):
			"""
			Conversion of a `logic_and`/`logic_or`.
			`unit`: distinguish between `logic_and` and `logic_or`
			"""
			new_chain = [];
			for arg in chain:
				new_arg = rec_subs(arg, not unit);
				if isinstance(new_arg, bool):
					if new_arg == unit:
						continue;
					else:
						return new_arg;
				new_chain.append(new_arg);
			return new_chain if new_chain else unit;
		
		def subs_relation (relation):
			"""
			Conversion of a `logic_relation`
			"""
			factors, mask = relation;
			sign = 1;
			for idx in factors:
				poly = subs_polys[idx];
				# Substitute poly if not already done in this step
				if not (done[0] >> idx) & 1:
					poly = poly(value);
					const = multi_constant(poly);
					# If poly becomes a constant, store its sign only
					if const is not None:
						poly = int(sgn(const));
						done[1] |= (1 << idx);
					subs_polys[idx] = poly;
					done[0] |= (1 << idx);
				# If it is a constant, contribute to the total sign
				if isinstance(poly, int):
					if poly == 0:
						sign = 0;
						break;
					elif poly < 0 and sign is not None:
						sign = -sign;
				else:
					sign = None;
			if sign is None:
				return relation;
			else:
				return bool(mask & sign_to_mask(sign));
		
		other = OptimizedFormula();
		other._tree = rec_subs(self._tree);
		other._vars = self._vars[1:];
		other._polys = subs_polys;
		other._consts = done[1];
		return other;
	
	def is_constant (self):
		"""Return whether the logical formula is constant true/false."""
		return isinstance(self._tree, bool);
	def __nonzero__ (self):
		return bool(self._tree);
	
	# Private members:
	# self._tree: the structure of the formula. It is built from the following
	#    compoments hierarchically:
	#    logic_and/logic_or: represented by lists (of lists). They alternately
	#       mean logic_and and logic_or, the outside list being a logic_and
	#       (which may be a one-element list if the outside operator would be
	#       a logic_or).
	#    logic_relation: represented by a (factors, mask) tuple as described
	#       below. The relation is reduced to zero on the right-hand-side,
	#       and the left-hand-side polynomial is factored.
	#       factors: list of indices of self._polys, meaning the factors of the
	#          polynomial.
	#       mask: the relational operator converted by relation_to_mask().
	#    logic_true/logic_false: represented by True/False.
	#       It is only possible on the root, otherwise they are optimized out.
	# self._polys: the list of the actual polynomials. It is actually computed
	#    by and borrowed from PolyFactorList, the only difference is that the
	#    variable order is reversed to simplify substitution.
	# self._vars: the tuple of variables of self._polys. It determines the
	#    order of substitution.
	# self._consts: bit mask showing which indices in self._polys are already
	#    constants. They are simplified to their sign (-1/0/+1), and they do
	#    not have to (and can) be substituted further.

