#
# Logical formulas
# License: GNU General Public License 3
# Uray M. János
# 2019-2020
#

#
# Logical basics
# --------------
#

class Logic:
	"""
	Base class of symbolic logical formulas.
	
	It has the following subclasses:
	`LogicConstant`: true or false
	`LogicRelation`: symbolic relational expression, e.g. x^2*y > (x-1)^2
	`LogicChain`: logican conjuction ("and") or disjunction ("or") of two or
		more logical formulas, which are themselves not `LogicChain`'s of the
		same type
	`LogicQuantified`: quantified (exists/forall) version of a logical formula
   
	Neither of these classes should be instantiated directly, instead the
	following functions and constants should be used:
	`logic_true`, `logic_false`: logical constants
	`logic()`: wrap a Sage expression or bool to Logic
	`logic_not()`: negation
	`logic_and()`, `logic_or()`: conjuction, disjunction
	`logic_implies()`: implication (A => B)
	`logic_forall()`, `logic_exists()`: quantified formula
	"""
	
	def __and__ (self, other):
		return logic_and(self, other);
	def __or__ (self, other):
		return logic_or(self, other);
	def __rand__ (self, other):
		return logic_and(other, self);
	def __ror__ (self, other):
		return logic_or(other, self);
	def __invert__ (self):
		raise NotImplementedError;
	def __nonzero__ (self):
		raise NotImplementedError;
	def __eq__ (self, other):
		raise NotImplementedError;
	def __ne__ (self, other):
		return not self.__eq__(other);
	
	def subs (self, *args, **kwargs):
		"""Substitute the given variables to the formula. Evaluate the formula if constant."""
		raise NotImplementedError;
	def __call__ (self, *args, **kwargs):
		return self.subs(*args, **kwargs);
	
	def is_constant (self):
		"""Return whether the logical formula is constant (i.e. true or false)."""
		return False;
	def operator (self):
		"""Return the operator of the logical formula, e.g. logic_and. It is None for a constant formula."""
		raise NotImplementedError;
	def operands (self):
		"""Return the operands of the logical formula, e.g. for a conjuction. It is an empty list for a constant formula.""";
		raise NotImplementedError;
	
	def variables (self):
		"""Return all variables in the formula as a sorted tuple."""
		raise NotImplementedError;
	def free_variables (self):
		"""Return all free variables in the formula as a sorted tuple, i.e. all variables not bound by quantifiers."""
		raise NotImplementedError;
	@staticmethod
	def _sort_vars (vars):
		from sage.symbolic.comparison import print_sorted;
		return tuple(reversed(print_sorted(vars)));
	
	def atoms (self):
		"""Iterate through all atomic subformulas of the formula in order of appearance."""
		raise NotImplementedError;
	
	def recursive_replace (self, replace):
		"""
		Replace all subformulas according to the given replace function, and
		return the modified new formula, keeping the original object unchanged.
		The replace function is called for all subformulas recursively, and it
		must return the modified (or the same) subformula. The children come
		after the parent, and always the new children are examined, never the
		replaced old ones.
		"""
		new = logic(replace(self));
		return new._recursive_replace_children(replace);
	def _recursive_replace_children (self, replace):
		return self;
	
	def to_prenex (self):
		"""Return an equivalent prenex formula.""";
		raise NotImplementedError;
	
	def top_quantifiers (self):
		"""Return a list of quantifiers which are on the top of the formula tree, or empty list if it is not a directly quantified formula."""
		return [];
	def quantifier_base (self):
		"""
		Return the formula inside the quantifiers returned by top_quantifier().
		
		For a prenex formula, quantifier_base() together with top_quantifiers()
		split the formula to quantifiers and the unquantified part.
		"""
		return self;


class LogicConstant (Logic):
	"""
	Locical constant, i.e. true or false.
	
	The class should not be instantiated directly, instead use the constants
	`logic_true` and `logic_false`.
	"""
	
	# self._bool: the logical True or False
	
	def __init__ (self, value):
		self._bool = bool(value);
	
	def subs (self, *args, **kwargs):
		return self;
	
	def __invert__ (self):
		return LogicConstant(not self._bool);
	
	def __nonzero__ (self):
		return self._bool;
	
	def __eq__ (self, other):
		if isinstance(other, LogicConstant):
			return self._bool == other._bool;
		elif isinstance(other, bool):
			return self._bool == other;
		else:
			return False;
	
	def is_constant (self):
		return True;
	def operator (self):
		return None;
	def operands (self):
		return [];
	
	def variables (self):
		return ();
	def free_variables (self):
		return ();
	
	def atoms (self):
		yield self;
	
	def to_prenex (self):
		return self;
	
	def __repr__ (self, child=False):
		return "logic_true" if self._bool else "logic_false";
	def __str__ (self):
		return "True" if self._bool else "False";

logic_true  = LogicConstant(True);
logic_false = LogicConstant(False);

class LogicRelation (Logic):
	"""
	Atomic logical formula: a symbolic equality or inequality. It wraps a Sage symbolic Expression of relational type.
	"""
	
	# self._rel: the equality or inequality as a symbolic expression
	
	def __init__ (self, rel):
		if rel.operator() not in (operator.eq, operator.gt, operator.ge, operator.lt, operator.le, operator.ne):
			raise ValueError("LogicRelation expects a symbolic equality/inequality");
		self._rel = rel;
	
	def subs (self, *args, **kwargs):
		return logic(self._rel.subs(*args, **kwargs));
	
	def __invert__ (self):
		return LogicRelation(self._rel.negation());
	
	def __eq__ (self, other):
		if isinstance(other, LogicRelation):
			return self._rel == other._rel;
		elif isinstance(other, Expression):
			return self._rel == other;
		else:
			return False;
	
	def operator (self):
		return self._rel.operator();
	def operands (self):
		return self._rel.operands();
	def lhs (self):
		return self._rel.lhs();
	def rhs (self):
		return self._rel.rhs();
	
	def variables (self):
		return self._rel.variables();
	def free_variables (self):
		return self._rel.free_variables();
	
	def atoms (self):
		yield self;
	
	def to_prenex (self):
		return self;
	
	def _symbolic_ (self, sring):
		"Convert to a symbolic expression"
		return self._rel;
	
	def __repr__ (self, child=False):
		str = repr(self._rel);
		if not child:
			str = "logic(" + str + ")";
		return str;
	def __str__ (self):
		return str(self._rel);

def logic (value):
	"""
	Wrap a Sage expression to a logical formula.
	
	It can accept a boolean value (`True` or `False`), a symbolic relational
	Expression or a logical formula (the latter is returned without change).
	A constant relational expression is evaluated to `logic_true` or `logic_false`.
	Note that all other `logic_*` functions apply `logic()` first to their
	parameters when they are expected to be logical formulas.
	"""
	if isinstance(value, Logic):
		return value;
	elif isinstance(value, bool):
		return LogicConstant(value);
	elif value.is_relational():
		try:
			return LogicConstant(value.operator()(AA(value.lhs()), AA(value.rhs())));
		except:
			return LogicRelation(value);
	else:
		raise ValueError("Unsupported value for logical formula");

def logic_not (op):
	"""
	Create the negation of the logical formula.
	"""
	# Note that there is no class for negated formulas, because negation
	# is always propagated through the formula tree.
	return ~logic(op);


#
# Boolean combination
# -------------------
#

class LogicChain (Logic):
	"""
	Logical "and" or "or" chain.
	
	This class should not be used directly, instead use the functions
	`logic_and()` and `logic_or()`.
	"""
	
	# self._ops: list of operands (two or more)
	# self._unit: the unit of the used logical operation:
	#    logic_true for "and" and logic_false for "or"
	
	def __init__ (self, ops, unit):
		assert len(ops) >= 2;
		self._ops = ops;
		self._unit = unit;
	
	def subs (self, *args, **kwargs):
		return _make_logic_chain(self._unit, (op.subs(*args, **kwargs) for op in self._ops));
	
	def __invert__ (self):
		return LogicChain([~op for op in self._ops], ~self._unit);
	
	def __eq__ (self, other):
		if not isinstance(other, LogicChain):
			return False;
		if self._unit != other._unit:
			return False;
		if len(self._ops) != len(other._ops):
			return False;
		for i in range(len(self._ops)):
			if self._ops[i] != other._ops[i]:
				return False;
		return True;
	
	def operator (self):
		return logic_and if self._unit else logic_or;
	def operands (self):
		return self._ops;
	
	def variables (self):
		s = set();
		for op in self._ops:
			s.update(op.variables());
		return self._sort_vars(s);
	
	def free_variables (self):
		s = set();
		for op in self._ops:
			s.update(op.free_variables());
		return self._sort_vars(s);
	
	def atoms (self):
		for op in self._ops:
			for atom in op.atoms():
				yield atom;
	
	def _recursive_replace_children (self, replace):
		new_ops = None;
		for i, op in enumerate(self._ops):
			new_op = op.recursive_replace(replace);
			if new_op is not op and new_ops is None:
				new_ops = self._ops[:i];
			if new_ops is not None:
				new_ops.append(new_op);
		if new_ops:
			return _make_logic_chain(self._unit, new_ops);
		else:
			return self;
	
	def to_prenex (self):
		# TODO quantifier variable clashes with outside variable
		quantifiers = [];
		new_ops = [];
		for op in self._ops:
			prenex_op = op.to_prenex();
			quantifiers.extend(prenex_op.top_quantifiers());
			new_ops.append(prenex_op.quantifier_base());
		prenex = _make_logic_chain(self._unit, new_ops);
		vars = set();
		for quan in reversed(quantifiers):
			if quan.variable() in vars:
				raise NotImplementedError("quantified variable names clash");
			vars.add(quan.variable());
			prenex = LogicQuantified(quan, prenex);
		return prenex;
	
	def __repr__ (self, child=False):
		name = "logic_and" if self._unit else "logic_or";
		return name + "(" + str.join(", ", (op.__repr__(True) for op in self._ops)) + ")";
	def __str__ (self):
		symbol = " /\\ " if self._unit else " \\/ ";
		def child_str (op):
			if isinstance(op, (LogicChain, LogicQuantified)):
				return "(" + str(op) + ")";
			else:
				return str(op);
		return str.join(symbol, (child_str(op) for op in self._ops));

# Simplify a logical chain, implementation of logic_and() and logic_or().
# Note that it does not simply return a LogicChain with the same parameters,
# but instead it tries to simplify the formula if possible. E.g. constants in
# the chain are applied (lazy evaluation), and children chains of the same type
# are merged together, e.g. (A /\ B) /\ C is merged to A /\ B /\ C.
def _make_logic_chain (unit, *ops):
	if len(ops) == 1 and hasattr(ops[0], "__iter__"):
		ops = ops[0];
	red_ops = [];
	for op in ops:
		if isinstance(op, LogicChain) and op._unit == unit:
			red_ops.extend(op._ops);
		else:
			op = logic(op);
			if op.is_constant():
				if op == unit:
					continue;
				else:
					return op;
			red_ops.append(op);
	if len(red_ops) >= 2:
		return LogicChain(red_ops, unit);
	elif red_ops:
		return red_ops[0];
	else:
		return unit;

def logic_and (*ops):
	"""Return the conjuction (logical "and" combination) of the subformulas."""
	return _make_logic_chain(logic_true, *ops);

def logic_or (*ops):
	"""Return the disjunction (logical "or" combination) of the subformulas."""
	return _make_logic_chain(logic_false, *ops);

def logic_implies (lhs, rhs):
	"""Return the implication ("A implies B") statement for parameters A and B."""
	# Note that there is no class for implication, instead it is converted
	# to a disjunction: A => B is converted to ~A \/ B.
	return logic_or(logic_not(lhs), rhs);


#
# Quantified formulas
# -------------------
#

class LogicQuantified (Logic):
	"""Quantified logical formula"""
	
	# self._quan: the quantifier (instance of the Quantifier class below) 
	# self._child: the subformula
	
	def __init__ (self, quan, child):
		if not isinstance(quan, Quantifier):
			raise TypeError("LogicQuantified: " + str(quan) + " is not a quantifier");
		self._quan = quan;
		self._child = logic(child);
	
	def subs (self, *args, **kwargs):
		# remove v from the arguments
		v = self._quan._var;
		vname = str(v);
		if vname in kwargs:
			del kwargs[vname];
		def delete_var (arg):
			if isinstance(arg, (list, tuple)):
				return list(a for a in (delete_var(a) for a in arg) if a is not None);
			elif isinstance(arg, dict):
				return {key : val for key, val in arg.items() if v not in key.variables()};
			elif isinstance(arg, Expression):
				return arg if v not in arg.lhs().variables() else None;
			else:
				raise ValueError("invalid argument for subs(): {}".format(arg));
		args = delete_var(args);
		return LogicQuantified(self._quan, self._child.subs(*args, **kwargs));
	
	def __eq__ (self, other):
		if not isinstance(other, LogicQuantified):
			return False;
		return self._quan == other._quan and self._child == other._child;
	
	def operator (self):
		return self._quan;
	def operands (self):
		return [self._child];
	
	def variables (self):
		vars = self._child.variables();
		if self._quan._var not in vars:
			vars = self._sort_vars(vars + (self._quan._var,));
		return vars;
	def free_variables (self):
		vars = self._child.free_variables();
		if self._quan._var in vars:
			vars = tuple(v for v in vars if v != self._quan._var);
		return vars;
	
	def atoms (self):
		return self._child.atoms();
	
	def _recursive_replace_children (self, replace):
		new_child = self._child.recursive_replace(replace);
		if new_child is not self._child:
			return LogicQuantified(self._quan, new_child);
		else:
			return self;
	
	def to_prenex (self):
		return LogicQuantified(self._quan, self._child.to_prenex());
	
	def top_quantifiers (self):
		return [self._quan] + self._child.top_quantifiers();
	def quantifier_base (self):
		return self._child.quantifier_base();
	
	def __repr__ (self, child=False):
		params = self._quan._repr_info();
		if not isinstance(params, tuple):
			params = (params,);
		name = params[0];
		params = params[1:] + (self._child.__repr__(True),);
		return name + "(" + str.join(", ", params) + ")";
	def __str__ (self):
		return str(self._quan) + ": " + str(self._child);


class Quantifier:
	"""Base class of quantifiers in logical formulas"""
	
	# self._var: the variable of quantification
	
	def __init__ (self, var):
		if not isinstance(var, Expression) or not var.is_symbol():
			raise ValueError("Quantifier: " + str(var) + " is not a variable");
		self._var = var;
	
	def variable (self):
		"""Return the variable in the quantifier."""
		return self._var;
	
	def evaluator (self):
		"""Return an object that can be evaluated lazily by giving some (or all) truth values for the elements."""
		raise NotImplementedError;
	
	def _repr_info (self):
		raise NotImplementedError;

class StdQuantifier (Quantifier):
	"""Universal or existential quantifier"""
	
	# self._unit: logic_true for "forall" and logic_false for "exists"
	
	def __init__ (self, unit, var):
		Quantifier.__init__(self, var);
		if not isinstance(unit, LogicConstant):
			raise TypeError("StdQuantifier expects either logic_true or logic_false as unit");
		self._unit = unit;
	
	def __eq__ (self, other):
		if not isinstance(other, StdQuantifier):
			return False;
		return self._unit == other._unit and self._var == self._var;
	
	def evaluator (self):
		return StdQuantifierEvaluator(self._unit);
	
	def _repr_info (self):
		return "logic_forall" if self._unit else "logic_exists";
	def __str__ (self):
		sign = "A" if self._unit else "E";
		return sign + " " + str(self._var);

class QuantifierEvaluator:
	"""Class for evaluating the quantifier by giving some truth values of the elements"""
	
	def value (self):
		"""Return True/False if the value is already known, or None otherwise."""
		raise NotImplementedError;
	
	def add (self, value, n=1):
		"""Add one or more boolean values to the evaluation. The parameter n gives the number of values, which can be Infinity."""
		raise NotImplementedError;
	
	def finish (self):
		"""Indicate that there are no more values that were added by add()."""
		raise NotImplementedError;

class StdQuantifierEvaluator (QuantifierEvaluator):
	def __init__ (self, unit):
		self._unit = bool(unit);
		self._value = None;
	def value (self):
		return self._value;
	def add (self, value, n=1):
		value = bool(value);
		if value != self._unit:
			self._value = value;
	def finish (self):
		if self._value is None:
			self._value = self._unit;


def logic_forall (var, formula):
	"""
	Create a universal quantified formula from `formula` by the variable `var`.
	
	E.g. logic_forall(x, x^2 + y^2 >= 0)
	"""
	return LogicQuantified(StdQuantifier(logic_true, var), formula);

def logic_exists (var, formula):
	"""
	Create an existential quantified formula from `formula` by the variable `var`.
	
	E.g. logic_exists(x, x^2 == y)
	"""
	return LogicQuantified(StdQuantifier(logic_false, var), formula);

