#
# Miscellanea for polynomials
# License: GNU General Public License 3
# Uray M. János
# 2019-2020
#

# For most of the functions in this file, 'polynomial' means a recursively
# defined multivariate polynomial (e.g. QQ[x][y]...)


class PolyFactorList:
	"""
	Class containing a list of irreducible multivariate polynomials in a given
	recursive polynomial field (e.g. QQ[x][y]...).
	Input can be any polynomial (reducible or irreducible) in either
	QQ[x][y]..., QQ[x,y,...] or SR. The class stores their factors,
	each polynomial as primitive and only once.
	"""
	
	def __init__ (self, ring):
		self._ring = ring;
		self._vars = self._fetch_ring_vars(self._ring)
		self._list = [];
	
	# Collect variables from a recursive polynomial ring (to a tuple).
	@staticmethod
	def _fetch_ring_vars (ring):
		vars = tuple();
		while isinstance(ring.gen(), Polynomial):
			vars = (ring.gen(),) + vars;
			ring = ring.base();
		return vars;
	
	def add (self, f):
		"""
		Add a new polynomial to the list. It is factored, and factors
		are stored only if not already present.
		"""
		f = self._ring(f);
		trace = [];
		const = multi_constant(f);
		while const is None:
			fact = f.factor();
			for g, m in fact:
				g = self._ring(g);
				try:
					idx = self._list.index(g);
				except ValueError:
					self._list.append(g);
					idx = len(self._list) - 1;
				for i in range(m):
					trace.append(idx);
			f = fact.unit();
			const = multi_constant(f);
		return (trace, const);
	
	def ring (self):
		"""The recursive polynomial ring"""
		return self._ring;
	def variables (self):
		"""Variables in a tuple"""
		return self._vars;
	def dimension (self):
		"""Dimension, i.e. the number of variables"""
		return len(self._vars);
	
	# Objects can be indexed like lists
	def __len__ (self):
		return len(self._list);
	def __getitem__ (self, i):
		return self._list[i];
	
	def filter (self, test):
		"""Remove polynomials that fail to satisfy the given test."""
		self._list = [poly for poly in self._list if test(poly)];


def subresultant (f, g, ret="subres"):
	"""
	Perform the subresultant algorithm.
	
	The return value depends on parameter `ret`:
	`ret` == "res": the resultant of `f` and `g`
	`ret` == "subres": the subresultants of `f` and `g` in a list,
		indexed by the subresultant order
	`ret` == "psc": the principal subresultant coefficients in a list,
		i.e. the k'th coefficient of each subres_k(`f`, `g`)
	"""
	n = min(f.degree(), g.degree()); # max subresultant index
	if f.degree() == g.degree():
		n -= 1;
	if ret == "subres":
		list = [f.parent()(0) for i in range(n+1)];
	elif ret == "psc":
		list = [f.base_ring()(0) for i in range(n+1)];
	if f == 0 or g == 0:
		if ret == "res":
			return f.parent()(0);
		else:
			return list;
	signs = None;
	if f.degree() < g.degree():
		f, g = g, f;
		if f.degree() & g.degree() & 1:
			signs = "even";
		if not (f.degree() | g.degree()) & 1:
			signs = "odd";
	fcont, gcont = content(f), content(g);
	f, g = f // fcont, g // gcont;
	cont = fcont ^ g.degree() * gcont ^ f.degree();
	prev_lc = 1;
	resultant = 1;
	while g != 0:
		degdiff = f.degree() - g.degree();
		f *= g.lc() ^ (degdiff + 1);
		f %= g;
		f //= -prev_lc * (-resultant) ^ degdiff;
		f, g = g, f;
		prev_lc = f.lc();
		if degdiff > 1:
			resultant = prev_lc ^ degdiff // resultant ^ (degdiff - 1);
		else:
			resultant = prev_lc ^ degdiff * resultant ^ (1 - degdiff);
		if ret == "subres":
			if degdiff > 1 or f.degree() == n:
				list[f.degree()] = f * resultant // prev_lc;
			if f.degree() > 0:
				list[f.degree() - 1] = g;
		elif ret == "psc":
			if f.degree() <= n:
				list[f.degree()] = resultant;
	if ret == "res":
		if f.degree() > 0:
			return 0;
		else:
			resultant *= cont;
			if signs == "even":
				resultant = -resultant;
			return resultant;
	else:
		if signs == "even":
			for k in range(0, n+1, 2):
				list[k] = -list[k];
		elif signs == "odd":
			for k in range(1, n+1, 2):
				list[k] = -list[k];
		return list;


#
# Misc functions
#

def multi_lc (f):
	"""
	Leading coefficient of `f`, treated recursively, i.e. a single number.
	
	E.g.: ((3*x + 1)*y^2 + (x^2 + 1)*y + 2*x^3 -> 3
	"""
	lc = f;
	while isinstance(lc, Polynomial):
		lc = lc.lc();
	return lc;

def multi_content (f):
	"""
	Content of `f` as a number, i.e. the gcd of all numbers in `f`.
	
	E.g. (4*x + 8)*y^2 + (2*x^2 + 4*x)*y + (6*x + 12) -> 2.
	Its sign depends on the sign of multi_lc(f).
	"""
	if not isinstance(f, Polynomial):
		return f;
	cont = 0;
	for c in f:
		ccont = multi_content(c);
		cont = gcd(cont, ccont);
	if multi_lc(f).sign() == -1:
		cont = -cont;
	return cont;

def multi_primitive (f):
	"""
	Primitive part of `f`, i.e. dividing it by multi_content(`f`).
	"""
	if f.is_zero():
		return f;
	cont = multi_content(f);
	if cont.is_one():
		return f;
	else:
		return f // cont;

def content (f):
	"""
	Content of `f`, i.e. the gcd of the direct coefficients.
	
	E.g. (4*x + 8)*y^2 + (2*x^2 + 4*x)*y + (6*x + 12) -> 2*x + 4.
	"""
	if not isinstance(f, Polynomial):
		return f;
	cont = 0;
	for c in f:
		cont = gcd(cont, c);
	if multi_lc(f).sign() == -1:
		cont = -cont;
	return cont;

def similar (f, g):
	"""
	Return if `f` and `g` are similar, i.e. one is constant multiple of the other.
	"""
	if f.degree() != g.degree():
		return False;
	if f.degree() <= 0:
		return True;
	flc = multi_lc(f);
	glc = multi_lc(g);
	if flc == glc:
		return f == g;
	elif flc == 1:
		return f*glc == g;
	elif glc == 1:
		return f == g*flc;
	else:
		return f*glc == g*flc;

def multi_constant (f):
	"""
	If `f` is a constant number, return it as a number type,
	otherwise return None.
	"""
	while isinstance(f, Polynomial):
		n = f.degree();
		if n > 0: return None;
		f = f.lc() if n == 0 else f.parent().base().zero();
	return f;


def find_algebraic_roots (f):
	"""
	Return the roots of an AA[x] polynomial as algebraic numbers (AA).
	It is the same as poly.roots(AA, multiplicities=False), but more efficient
	for lower degrees (and polynomials with lower degree factors).
	"""
	from heapq import merge;
	roots = [];
	for g, _ in f.factor() if f else []:
		d = g.degree();
		new_roots = [];
		if d == 1:
			b, a = g.coefficients(sparse=False);
			new_roots.append(AA(-b/a));
		elif d == 2:
			c, b, a = g.coefficients(sparse=False);
			d = b^2 - 4*a*c;
			s = sign(d);
			if s > 0:
				d = AA(sqrt(AA(d)));
				p2a = 1/(2*a);
				new_roots.append(AA((-b - d)*p2a));
				new_roots.append(AA((-b + d)*p2a));
				if p2a < 0: new_roots.reverse();
			elif s == 0:
				new_roots.append(AA(b/(2*a)));
		else:
			new_roots = g.roots(AA, multiplicities=False);
		if roots:
			roots = list(merge(roots, new_roots));
		else:
			roots = new_roots;
	return roots;

