#
# QEPCAD wrapper
# License: GNU General Public License 3
# Uray M. János
# 2019-2020
#

load("logic.sage");

#
# Execute the QEPCAD command.
# input: the input prenex formula (instance of Logic)
# vars: list of variables to indicate order
# n_free_vars: the first so many elements of vars are free variables
# memcells: number of memory cells to be allocated by QEPCAD.
#    Default is 2000000.
#
def use_qepcad (input, vars, n_free_vars, memcells=None):
	import pexpect;
	import re;
	command = "qepcad";
	if memcells is None:
		memcells = 4000000;
	if memcells is not None:
		command += " +N{}".format(memcells);
	qepcad = pexpect.spawn(command, timeout=None);
	qepcad.expect("\r\nEnter .*:\r\n");
	qepcad.sendline("[ cylindrical_algebraic_decomposition() in Sage ]");
	qepcad.expect("\r\nEnter .*:\r\n");
	qepcad.sendline("(" + ",".join([str(v) for v in vars]) + ")");
	qepcad.expect("\r\nEnter .*:\r\n");
	qepcad.sendline(str(n_free_vars));
	qepcad.expect("\r\nEnter .*:\r\n");
	qepcad.sendline(input + ".");
	ok = qepcad.expect(["\r\n([^\r\n]*) >\r\n", "\r\nError ([^\r\n]*)\r\n"]) == 0;
	if not ok:
		raise SyntaxError("QEPCAD reports for " + input + ": " + qepcad.match.group(1));
	qepcad.sendline("finish");
	qepcad.expect(pexpect.EOF);
	match = re.search("\r\nAn equivalent quantifier-free formula:(.*)\r\n=+", qepcad.before, re.DOTALL);
	if not match:
		raise SyntaxError("Cannot parse QEPCAD output");
	return match.group(1).strip();


#
# formula -> QEPCAD
# -----------------
#

def formula_to_qepcad (form):
	def expr_to_qepcad (expr):
		for v in expr.variables():
			if not expr.is_polynomial(v):
				raise ValueError("invalid expression {} for QEPCAD".format(expr));
		s = str(expr);
		s = s.replace("*", " ");
		return s;
	op = form.operator();
	if op is None:
		return "TRUE" if form else "FALSE";
	elif op == operator.eq:
		return expr_to_qepcad(form.lhs()) + " = " + expr_to_qepcad(form.rhs());
	elif op == operator.ne:
		return expr_to_qepcad(form.lhs()) + " /= " + expr_to_qepcad(form.rhs());
	elif op == operator.lt:
		return expr_to_qepcad(form.lhs()) + " < " + expr_to_qepcad(form.rhs());
	elif op == operator.gt:
		return expr_to_qepcad(form.lhs()) + " > " + expr_to_qepcad(form.rhs());
	elif op == operator.le:
		return expr_to_qepcad(form.lhs()) + " <= " + expr_to_qepcad(form.rhs());
	elif op == operator.ge:
		return expr_to_qepcad(form.lhs()) + " >= " + expr_to_qepcad(form.rhs());
	elif op in (logic_and, logic_or):
		if any(isinstance(arg.operator(), Quantifier) for arg in form.operands()):
			raise ValueError("QEPCAD accepts only prenex formula");
		symbol = " /\\ " if op == logic_and else " \\/ ";
		return "[" + str.join(symbol, (formula_to_qepcad(arg) for arg in form.operands())) + "]";
	elif isinstance(op, Quantifier):
		return "(" + str(op) + ")" + formula_to_qepcad(form.operands()[0]);
	else:
		raise ValueError("invalid formula {} for QEPCAD".format(form));


#
# QEPCAD -> formula
# -----------------
#

def qepcad_to_formula (str):
	def formula_read (str, i):
		while i < len(str) and str[i].isspace():
			i += 1;
		if i == len(str):
			raise QepcadSyntaxError(str, i);
		c = str[i];
		if c == "[":
			i += 1;
			(form, i) = formula_read(str, i);
			if i == len(str) or str[i] != "]":
				raise QepcadSyntaxError(str, i, "expected ]");
			i += 1;
		elif c == "~":
			i += 1;
			(form, i) = formula_read(str, i);
			form = logic_not(form);
		else:
			(form, i) = atomic_read(str, i);
		
		logic_op_chars = "/\\<=>";
		while i < len(str):
			c = str[i];
			if c.isspace():
				i += 1;
			elif c in logic_op_chars:
				op = c;
				i += 1;
				while i < len(str) and str[i] in logic_op_chars:
					op += str[i];
					i += 1;
				if op == "/\\":
					(right, i) = formula_read(str, i);
					form = logic_and(form, right);
				elif op == "\\/":
					(right, i) = formula_read(str, i);
					form = logic_or(form, right);
				elif op == "<==":
					(right, i) = formula_read(str, i);
					form = logic_implies(right, form);
				elif op == "==>":
					(right, i) = formula_read(str, i);
					form = logic_implies(form, right);
				elif op == "<==>":
					i -= len(op);
					raise QepcadSyntaxError(str, i, "sorry, equivalence is not implemented");
				else:
					i -= len(op);
					raise QepcadSyntaxError(str, i, "invalid logical operator " + op);
			else:
				break;
		return (form, i);
	
	def atomic_read (str, i):
		while i < len(str) and str[i].isspace():
			i += 1;
		if i == len(str):
			raise QepcadSyntaxError(str, i);
		if str[i:i+4] == "TRUE":
			return (logic_true, i+4);
		elif str[i:i+5] == "FALSE":
			return (logic_false, i+5);
		(left, i) = expression_read(str, i);
		while i < len(str) and str[i].isspace():
			i += 1;
		rel_op_chars = "<=>/";
		op = "";
		while i < len(str) and str[i] in rel_op_chars:
			op += str[i];
			i += 1;
		if op == "=":
			(right, i) = expression_read(str, i);
			rel = (left == right);
		elif op == "/=":
			(right, i) = expression_read(str, i);
			rel = (left != right);
		elif op == "<":
			(right, i) = expression_read(str, i);
			rel = (left < right);
		elif op == ">":
			(right, i) = expression_read(str, i);
			rel = (left > right);
		elif op == "<=":
			(right, i) = expression_read(str, i);
			rel = (left <= right);
		elif op == ">=":
			(right, i) = expression_read(str, i);
			rel = (left >= right);
		elif op:
			i -= len(op);
			raise QepcadSyntaxError(str, i, "invalid relational operator " + op);
		else:
			raise QepcadSyntaxError(str, i, "relational operator expected");
		return (logic(rel), i);
	
	def expression_read (str, i, prec=0):
		while i < len(str) and str[i].isspace():
			i += 1;
		if i == len(str):
			raise QepcadSyntaxError(str, i);
		c = str[i];
		if c == "(":
			i += 1;
			(expr, i) = expression_read(str, i, 0);
			if i == len(str) or str[i] != ")":
				raise QepcadSyntaxError(str, i, "expected )");
			i += 1;
		elif c == "+":
			i += 1;
			(expr, i) = expression_read(str, i, 1);
			expr = +expr;
		elif c == "-":
			i += 1;
			(expr, i) = expression_read(str, i, 1);
			expr = -expr;
		elif c.isdigit():
			n = c;
			i += 1;
			while i < len(str) and str[i].isdigit():
				n += str[i];
				i += 1;
			expr = SR(n);
		elif c.isalpha():
			v = c;
			i += 1;
			while i < len(str) and str[i].isalpha():
				v += str[i];
				i += 1;
			expr = SR(v);
		else:
			raise QepcadSyntaxError(str, i);
		
		while i < len(str):
			c = str[i];
			if c.isspace():
				i += 1;
			elif c == "+":
				if prec >= 1: break;
				i += 1;
				(right, i) = expression_read(str, i, 1);
				expr += right;
			elif c == "-":
				if prec >= 1: break;
				i += 1;
				(right, i) = expression_read(str, i, 1);
				expr -= right;
			elif c == "*":
				if prec >= 2: break;
				i += 1;
				(right, i) = expression_read(str, i, 2);
				expr *= right;
			elif c == "/" and not (i+1 < len(str) and str[i+1] in "\\="):
				if prec >= 2: break;
				i += 1;
				(right, i) = expression_read(str, i, 2);
				expr /= right;
			elif c == "^":
				if prec >= 3: break;
				i += 1;
				(right, i) = expression_read(str, i, 3);
				expr ^= right;
			elif c == "(" or c.isdigit() or c.isalpha():
				# unmarked multiplication
				if prec >= 2: break;
				(right, i) = expression_read(str, i, 2);
				expr *= right;
			else:
				break;
		return (expr, i);
	
	(form, i) = formula_read(str, 0);
	if i < len(str):
		raise QepcadSyntaxError(str, i);
	return form;

class QepcadSyntaxError (SyntaxError):
	def __init__ (self, str, i, info = None):
		if info:
			info += " at ";
		else:
			info = "";
		next_part = str[i:i+10];
		if i + 10 < len(str):
			info += "\"" + next_part + "\"... in QEPCAD-string";
		elif next_part:
			info += "\"" + next_part + "\" in QEPCAD-string";
		else:
			info += "end of QEPCAD-string";
		SyntaxError.__init__(self, info);

