#
# Recursifier
# License: GNU General Public License 3
# Uray M. János
# 2019-2020
#

from sage.symbolic.operators import add_vararg, mul_vararg;

load("logic.sage");

class Recurrence:
	"""
	Class representing a recurrence relation, created from a recurrence equation
	"""
	
	# self._base : the expression whose recurrence relation is defined.
	#    It consists of a symbolic function whose arguments are symbolic
	#    variables, or they may also be function symbols like f(k), where k is
	#    also an argument of self._base (e.g. as in sum(f(k), k, a, n)).
	# self._n : the recursion parameter ("n"), an argument of self._base
	# self._n_idx : the index of the recursion parameter ("n") in the argument list
	# self._shifted : the shifted version of self._base, i.e. the right-hand
	#    side of the recurrence relation.
	# self._order : the order of the recursion
	# self._min : the lowest value of self._n where the function is defined
	# self._init: list of initial values from self._min of length self._order
	# self._bound : the variables of self._base which are bound to a symbolic
	#    function variable (e.g. k in the example).
	
	def __init__ (self, eqs, min=None):
		if not isinstance(eqs, list):
			eqs = [eqs];
		if not eqs:
			raise ValueError("Empty list for Recurrence");
		func = self._get_common_function_pattern(eqs);
		rec_idx = self._find_the_recurrence(eqs, func);
		self._n_idx = self._find_the_different_arg(eqs, func);
		rec = eqs[rec_idx];
		del eqs[rec_idx];
		self._order = self._calc_recurrence_order(rec, func, self._n_idx);
		rec = self._adjust_recurrence(rec, self._n_idx);
		self._base = rec.left();
		self._n = self._base.operands()[self._n_idx];
		self._check_arguments();
		self._shifted = rec.right().subs(self._n == self._n + self._order)
		init = self._initial_values_eqs_to_dict(eqs);
		self._min = ZZ(min) if min is not None else None;
		self._collect_initial_values(init);
	
	@staticmethod
	def _is_function_expr (expr):
		return expr.operator() not in (None, add_vararg, mul_vararg);
	
	@staticmethod
	def _get_common_function_pattern (eqs):
		func_op = None;
		for eq in eqs:
			if not isinstance(eq, Expression) or eq.operator() is not operator.eq:
				raise TypeError("{} is not an equation".format(eq));
			left = eq.left();
			if not Recurrence._is_function_expr(left):
				raise ValueError("{} is not a function".format(left));
			if func_op is None:
				func_op = left.operator();
			elif func_op != left.operator():
				raise ValueError("Left sides are not all the same functions");
		try:
			nargs = func_op.number_of_arguments();
		except:
			nargs = eqs[0].left().number_of_operands();
		return func_op(*(SR.wild(i) for i in range(nargs)));
	
	@staticmethod
	def _find_the_recurrence (eqs, func):
		rec_idx = None;
		for i, eq in enumerate(eqs):
			if eq.right().has(func):
				if rec_idx is not None:
					raise ValueError("Multiple recurrence relations given for {}()".format(func.operator()));
				rec_idx = i;
		if rec_idx is None:
			if len(eqs) == 1:
				raise ValueError("{} is not a recurrence relation".format(eqs[0]));
			else:
				raise ValueError("No recurrence relation given for {}()".format(func.operator()));
		return rec_idx;
	
	@staticmethod
	def _find_the_different_arg (eqs, func):
		first = eqs[0].left();
		first_args = first.operands();
		diff_idx = None;
		for eq in eqs:
			for match in eq.find(func):
				if match.number_of_operands() != first.number_of_operands():
					raise ValueError("Occurences of {}() must have the same number of arguments".format(first.operator()));
				for i, arg in enumerate(match.operands()):
					if arg != first_args[i]:
						if diff_idx is not None and diff_idx != i:
							raise ValueError("Occurences of {}() may differ only in the recursion parameter".format(first.operator()));
						diff_idx = i;
		if diff_idx is None:
			raise ValueError("Occurrences of {} must differ in the recursion parameter".format(first));
		return diff_idx;
	
	@staticmethod
	def _calc_recurrence_order (rec, func, n_idx):
		left = rec.left();
		left_n = left.operands()[n_idx];
		order = 0;
		for match in rec.right().find(func):
			match_n = match.operands()[n_idx];
			try:
				shift = ZZ(left_n - match_n);
			except:
				raise ValueError("{} is not a shifted version of {}".format(match, left));
			if shift <= 0:
				raise ValueError("{} has higher or equal index than the left side".format(match));
			if shift > order:
				order = shift;
		return order;
	
	@staticmethod
	def _adjust_recurrence (rec, n_idx):
		left_n = rec.left().operands()[n_idx];
		n = left_n.default_variable();
		try:
			shift = ZZ(left_n - n);
		except:
			raise ValueError("Parameter {} of {} is not a variable".format(left_n, rec.left()));
		return rec.subs(n == n - shift);
	
	def _check_arguments (self):
		func_args = self._base.operands();
		self._bound = set();
		for i, arg in enumerate(func_args):
			if i != self._n_idx and self._n in arg.variables():
				raise ValueError("Other parameters of {} must not depend on the recursion parameter '{}'".format(self._base, self._n));
			if not arg.is_symbol():
				if self._is_function_expr(arg) and i != self._n_idx:
					for a in arg.operands():
						if not a.is_symbol():
							raise ValueError("Parameter of {} must be a variable".format(arg));
						if a not in func_args:
							raise ValueError("Parameter of {} must be also an independent parameter of {}".format(arg, self._base));
						self._bound.add(a);
				else:
					raise ValueError("Parameter of {} must be a variable".format(self._base));
	
	def _initial_values_eqs_to_dict (self, eqs):
		init = dict();
		for eq in eqs:
			n_arg = eq.left().operands()[self._n_idx];
			try:
				k = ZZ(n_arg);
			except:
				if self._n in n_arg.variables():
					raise ValueError("{} is not a recurrence relation or initial value definition".format(eq));
				else:
					raise ValueError("{} is not an integer constant in {}".format(n_arg, eq.left()));
			if self._n in eq.right().variables():
				raise ValueError("Right side of {} depends on '{}'".format(eq, self._n));
			if k in init:
				raise ValueError("{} has multiple initial values".format(eq.left()));
			init[k] = eq.right();
		return init;
	
	def _collect_initial_values (self, init):
		if init:
			n0 = min(k for k in init);
			if self._min is None:
				self._min = n0;
			elif n0 < self._min:
				raise ValueError("{} is below min={}".format(self._uneval_init(n0), self._min));
		else:
			self._init = None;
			return;
		n_max = max(k for k in init);
		if n_max > self._min + self._order - 1:
			raise ValueError("{} is too high for initial value".format(self._uneval_init(n_max)));
		self._init = [self._base.subs(self._n == self._min + i).unhold() for i in range(self._order)];
		for k, value in enumerate(self._init, start=self._min):
			if k in init:
				if init[k] != value and value.operator() != self._base.operator():
					raise ValueError("{} == {} is wrong (it should be {})".format(self._uneval_init(k), init[k], value));
				self._init[k - self._min] = init[k];
	
	def __repr__ (self):
		s = "Recurrence(";
		if self._init: s += "[";
		s += "{} == {}".format(self._base.subs(self._n == self._n + self._order), self._shifted);
		if self._init:
			for k, value in enumerate(self._init, start=self._min):
				s += ", {} == {}".format(self._uneval_init(k), value);
			s += "]";
		s += ")";
		return s;
	
	def __str__ (self):
		s = str(self._base);
		s += " " + "-->";
		if self._order > 1:
			s += "(" + str(self._order) + ")";
		s += " " + str(self._shifted);
		if self._init:
			for k, value in enumerate(self._init, start=self._min):
				s += ", {} -> {}".format(k, value);
		return s;
	
	def _uneval_init (self, k):
		args = self._base.operands();
		args[self._n_idx] = k;
		return self._base.operator()(*args, hold=True);
	
	def _subs_map_for (self, expr):
		if expr.operator() != self._base.operator():
			return;
		args = expr.operands();
		bargs = self._base.operands();
		if len(args) != len(bargs):
			return;
		map = dict();
		for arg, barg in zip(args, bargs):
			if barg.is_symbol():
				map[barg] = arg;
		for arg, barg in zip(args, bargs):
			if barg.operator():
				to_wild = [map[ba] == SR.wild(i) for i, ba in enumerate(barg.operands())];
				wilds = [eq.right() for eq in to_wild];
				map[barg.operator()(*wilds)] = arg.subs(to_wild);
		for bvar in self._bound:
			if not map[bvar].is_symbol():
				return;
		return map;


known_recurrences = None;


def add_recurrence (recurrence, min=None):
	"""Add a new global recurrence relation for Recursifier."""
	init_known_recurrences();
	if not isinstance(recurrence, Recurrence):
		recurrence = Recurrence(recurrence, min=min);
	func = recurrence._base.operator();
	if func not in known_recurrences:
		known_recurrences[func] = [];
	known_recurrences[func].append(recurrence);

# Initialize the built-in recurrence relations on first call.
def init_known_recurrences ():
	global known_recurrences;
	if known_recurrences is not None:
		return;
	n = SR.symbol("n");
	k = SR.symbol("k");
	a = SR.symbol("a");
	f = sage.symbolic.function_factory.function("f", nargs=1);
	known_recurrences = dict();
	add_recurrence(factorial(n) == n * factorial(n-1), min=0);
	add_recurrence(pow(a, n+1) == a * pow(a, n));
	add_recurrence(sum(f(k), k, a, n+1, hold=True) == sum(f(k), k, a, n, hold=True) + f(n+1));
	add_recurrence(product(f(k), k, a, n+1, hold=True) == product(f(k), k, a, n, hold=True) * f(n+1));
	#add_recurrence(fib(n+2) == fib(n+1) + fib(n));
	#add_recurrence(fibpol(n+2, x) == x*fibpol(n+1, x) + fibpol(n, x));
	#add_recurrence(expint(n+1, x) == (exp(-x) - x*expint(n, x)) / n);
	#add_recurrence(gammapol(n+1, x) == gammapol(n, x) * n + x^n * exp(-x));


class Recursifier:
	"""
	A class to convert explicit functions of an integral variable
	to recursive components.
	
	EXAMPLES::
	
		sage:  var('n,x');
		sage:  rec = Recursifier(n);
		sage:  ineq = rec.recursify(1 + n*x <= (x + 1)^n); ineq
		a*b + 1 <= c
		sage:  for v in rec.variables():
		....:      print v, "->", rec.shift(v);
		a -> a + 1
		b -> b
		c -> (b + 1)*c
		sage:  rec.shift(ineq);
		(a + 1)*b + 1 <= (b + 1)*c
	"""
	
	# self._n : the discrete parameter ("n") of the recursion (its generated symbol)
	# self._min : the initial value of the recursion (integer)
	# self._last_symbol : the name of the last generated symbol
	# self._expr_map : dictionary from recursified sub-expressions to the
	#    corresponding generated symbol (or sometimes to combination of symbols)
	# self._meaning : dictionary from symbols to their meanings, which is
	#    mostly the reverse of self._expr_map
	# self._shift : dictionary from each generated symbol to their shifted value
	# self._subs : dictionary from the unneccessary symbols to their meaning.
	#    They were created first as symbols, then later it turned out that they
	#    can be replaced by an expression with other symbols.
	# self._recurrences : local recurrence relations in addition to the global
	#    known_recurrences
	# self._functions : symbolic functions without recurrence relations
	
	def __init__ (self, n, min=0):
		"""
		Create a new Recursifier class.
		
		A class caches the already recursified expressions, so it is better to
		reuse the same Recursifier object than to always create a new one.
		
		`n` : the recursion variable
		`min` : where to start the recursion (this value may increase later if
			a recursive function is detected with higher initial value)
		"""
		if not n.is_symbol():
			raise TypeError("{} is not a symbolic variable".format(n));
		self._min = ZZ(min);
		self._last_symbol = "";
		self._expr_map = dict();
		self._meaning = dict();
		self._shift = dict();
		self._subs = dict();
		self._n = self._create_symbol();
		self._expr_map[n] = self._n;
		self._meaning[self._n] = n;
		self._shift[self._n] = self._n + 1;
		init_known_recurrences();
		self._recurrences = dict();
		self._functions = set();
	
	def add_recurrence (self, recurrence, min=None):
		"""Add a new recurrence relation."""
		if not isinstance(recurrence, Recurrence):
			recurrence = Recurrence(recurrence, min=min);
		func = recurrence._base.operator();
		if func not in known_recurrences:
			self._recurrences[func] = [];
		self._recurrences[func].append(recurrence);
	
	def add_symbolic_function (self, func):
		"""
		Add a new symbolic function with undefined recurrence relation.
		These functions may be treated as arbitrary sequences.
		"""
		if isinstance(func, Expression):
			func = func.operator();
		try:
			func.number_of_arguments();
		except:
			raise TypeError("{} is not a symbolic function".format(func));
		self._functions.add(func);
	
	def recursify (self, input):
		"""
		Convert an expression to its recursified form.
		
		`input` can be an Expression, Logic, or list or tuple of them.
		It is better to call the function once with a list than to call
		the function separately for each element.
		"""
		output = self._recursify(input);
		output = self.optimize(output);
		return output;
	
	# Recursify the input recursively.
	def _recursify (self, input):
		if isinstance(input, list):
			return [self._recursify(elem) for elem in input];
		if isinstance(input, tuple):
			return (self._recursify(elem) for elem in input);
		if isinstance(input, Logic):
			return input.recursive_replace(lambda f : self._rec_expr(SR(f)) if isinstance(f, LogicRelation) else f);
		if input in SR:
			return self._rec_expr(input);
		raise TypeError("recursify() got invalid type");
	
	# Recursify the expression recursively.
	def _rec_expr (self, expr):
		if expr.parent() is not SR:
			expr = SR(expr);
		if expr in self._expr_map:
			return self._expr_map[expr];
		op = expr.operator();
		args = expr.operands();
		if op is None:
			if expr.is_numeric() and expr.is_exact():
				rec = expr;
			elif expr.is_symbol():
				rec = self._create_constant(expr);
		elif expr.is_relational() or op in (add_vararg, mul_vararg):
			rec = op(*(self._rec_expr(arg) for arg in args));
		elif op is operator.pow and args[1] in QQ:
			rec = pow(self._rec_expr(args[0]), args[1]);
		elif op in (min_symbolic, max_symbolic, sage.functions.other.abs):
			rec = op(*(self._rec_expr(arg) for arg in args));
		elif op in self._recurrences:
			rec = self._rec_funcs(expr, self._recurrences[op]);
		elif op in self._functions:
			rec = self._rec_sym_func(expr);
		elif op in known_recurrences:
			rec = self._rec_funcs(expr, known_recurrences[op]);
		else:
			raise TypeError("Expressions with {} are not supported by Recursifier".format(op));
		if rec is not None:
			return rec;
		else:
			raise TypeError("Expressions like {} are not supported by Recursifier".format(expr));
	
	def _rec_funcs (self, expr, recurs):
		for recur in recurs:
			map = recur._subs_map_for(expr);
			if map is not None:
				return self._rec_func(expr, recur, map);
	
	# Recursify a known recursive function.
	def _rec_func (self, expr, recur, map=None):
		if map is None:
			map = recur._subs_map_for(expr);
			if map is None:
				return;
		n = self.n_var();
		for v in map:
			if v != recur._n and n in map[v].variables():
				return;
		nv = map[recur._n];
		w0, w1 = SR.wild(0), SR.wild(1);
		linear = (
			nv.match(w1*n + w0) or
			nv.match(w1*n) or
			nv.match(n + w0) or
			({} if nv == n else {w0: nv, w1: 0})
		);
		try:
			c0 = ZZ(linear[w0] if w0 in linear else 0);
			c1 = ZZ(linear[w1] if w1 in linear else 1);
		except:
			return;
		if c1 < 0:
			return;
		if c1 == 0:
			return self._rec_func_const(expr, recur, map, c0);
		else:
			return self._rec_func_linear(expr, recur, map, c0, c1);
	
	# Recursify a known recursive function with constant parameter (c).
	def _rec_func_const (self, expr, recur, map, c):
		if recur._min is not None:
			if c < recur._min:
				raise ValueError("{} is undefined".format(expr));
			if c >= recur._min + recur._order:
				map[recur._n] = c - recur._order;
				value = recur._shifted.subs(map).unhold();
				return self._rec_expr(value);
		if recur._init:
			value = recur._init[c - recur._min];
			value = value.subs(map);
		else:
			value = expr;
		value = value.unhold();
		if expr.is_trivially_equal(value):
			return self._create_constant(expr);
		else:
			return self._rec_expr(value);
	
	# Recursify a known recursive function with linear parameter (c1*n + c0).
	def _rec_func_linear (self, expr, recur, map, c0, c1):
		if recur._min is not None:
			min = ceil((recur._min - c0) / c1);
			if min > self._min:
				self._min = min;
		n = self.n_var();
		images = [];
		for i in range(recur._order + c1):
			old_img = self._expr_map[expr] if expr in self._expr_map else None;
			if i < recur._order:
				if old_img is None:
					img = self._create_symbol();
				else:
					img = old_img;
			else:
				img = self._rec_expr(recur._shifted.subs(map));
				map[recur._n] += 1;
				if old_img is not None and old_img.is_symbol():
					self._subs_symbol(old_img, img);
			if old_img is None:
				self._expr_map[expr] = img;
				if img.is_symbol():
					self._meaning[img] = expr;
			images.append(img);
			expr = expr.subs(c1*n == c1*n + 1);
		for i in range(recur._order):
			if images[i].is_symbol():
				self._shift[images[i]] = images[i + c1];
		return images[0];
	
	# Recursify a symbolic function with undefined recurrence.
	def _rec_sym_func (self, expr):
		sym = self._create_symbol();
		self._expr_map[expr] = sym;
		self._meaning[sym] = expr;
		return sym;
	
	# Shift a symbolic function with undefined recurrence.
	# This usually creates a new variable.
	def _shift_sym_func (self, sym):
		expr = self._meaning[sym];
		n = self.n_var();
		shifted = expr.subs(n == n + 1);
		ssym = self._rec_expr(shifted);
		self._shift[sym] = ssym;
		return ssym;
	
	# Create a new symbol.
	def _create_symbol (self):
		add = True;
		next = "";
		for ch in reversed(self._last_symbol):
			if add:
				if ch != "z":
					ch = chr(ord(ch) + 1);
					add = False;
				else:
					ch = "a";
			next = ch + next;
		if add:
			next = "a" + next;
		self._last_symbol = next;
		return SR.symbol(next);
	
	# Create a new symbol for a constant expression.
	def _create_constant (self, expr):
		sym = self._create_symbol();
		self._expr_map[expr] = sym;
		self._meaning[sym] = expr;
		self._shift[sym] = sym;
		return sym;
	
	def _subs_symbol (self, sym, value):
		subs = (sym == value);
		def subs_in_map (map):
			for key, value in map.iteritems():
				if sym in value.variables():
					map[key] = value.subs(subs);
		subs_in_map(self._expr_map);
		subs_in_map(self._shift);
		subs_in_map(self._subs);
		self._subs[sym] = value;
		if sym in self._shift and self._shift[sym].is_symbol():
			self._subs_symbol(self._shift[sym], value.subs(self._shift));
	
	def optimize (self, expr):
		return expr.subs(self._subs);
	
	def n_var (self):
		"""Return the recursion parameter ("n")."""
		return self._meaning[self._n];
	
	def variables (self):
		"""Return all created symbols."""
		from sage.symbolic.comparison import print_sorted;
		vs = [];
		for v in self._meaning:
			if v not in self._subs:
				vs.append(v);
		return reversed(print_sorted(vs));
	
	def meaning (self, sym):
		"""Return the meaning of a given symbol introduced by recursification."""
		return self._meaning[sym];
	
	def min (self):
		"""Return the minimum value of the discrete parameter ("n")."""
		return self._min;
	
	def eval (self, expr, k):
		"""Evaluate `expr` at a constant value, where `expr` was returned by  recursify()."""
		n = self.n_var();
		if k < self._min:
			raise ValueError("Recursion is not defined at {}={}".format(n, k));
		k_map = dict();
		for sym in expr.variables():
			value = self._meaning[sym].subs(n == k);
			k_map[sym] = self._rec_expr(value.unhold());
		for sym, value in k_map.iteritems():
			k_map[sym] = self.optimize(value);
		return expr.subs(k_map);
	
	def shift (self, expr):
		"""Return the shifted version of `expr`, where `expr` was returned by  recursify()."""
		for sym in expr.variables():
			if sym not in self._meaning:
				raise ValueError("{} is not a recursified expression".format(expr));
			if sym not in self._shift:
				self._shift_sym_func(sym);
		return expr.subs(self._shift);

#var("j,k,n,x");
#function("f", nargs=1);
#function("fib", nargs=1);
#function("fibpol", nargs=2);
#add_recurrence([fib(n+1) == fib(n) + fib(n-1), fib(0) == 0, fib(1) == 1]);
#rec = Recursifier(n, min=0);
#rec.add_recurrence([fibpol(n+1, x) == x*fibpol(n, x) + fibpol(n-1, x), fibpol(0, x) == 0, fibpol(1, x) == 1]);
#rec.add_symbolic_function(f);
##input = factorial(2*n) + f(2)*f(2*n) + f(n+1);
##input = factorial(2*n);
##input = fib(2*n) + 2*fib(2*n+1)^2 - 3*fib(2*n+2);
##input = product(fibpol(2*k, x), k, 1, n);
##input = 2*factorial(n+1) + n*2^(n-2) - n^2*7/2 + 7/3;
##input = 2*sum(3*sum(factorial(j), j, 1, k, hold=True), k, 0, n, hold=True);
##input = logic_implies(n >= 2, factorial(n) * sum((-1)^k / factorial(k), k, 0, n) > 0);
##input = logic_implies(n >= 2, fib(n) >= ((1 + sqrt(5))/2)^(n-2));
#
#from sage.symbolic.comparison import print_sorted
#
#output = rec.recursify(input);
#print input;
#print output;
#shifted = output;
#vs = set(output.variables());
#for i in range(3):
	#shifted = rec.shift(shifted);
	#vs |= set(shifted.variables());
	#print shifted;
#for i in range(rec.min(), rec.min() + 4):
	#const = rec.eval(output, i);
	#vs |= set(const.variables());
	#print i, "->", const;
#for v in reversed(print_sorted(vs)):
	#print v, ":", rec.meaning(v);

